@@ -3,28 +3,36 @@ import copy
 import json
 from sklearn.pipeline import make_pipeline
 from river import compose
+from typing import Union, List, Dict, Generator, Any
 
 from spotPython.utils.convert import class_for_name
 from spotPython.utils.transform import transform_hyper_parameter_values
 
 
-def generate_one_config_from_var_dict(var_dict, fun_control) -> dict:
+def generate_one_config_from_var_dict(
+    var_dict: Dict[str, np.ndarray], fun_control: Dict[str, Union[List[str], str]]
+) -> Generator[Dict[str, Union[int, float]], None, None]:
     """Generate one configuration from a dictionary of variables (as a generator).
-    This function takes a dictionary of variables as input arguments and returns a dictionary
-    with the values from the arrays in the dictionary.
+
+    This function takes a dictionary of variables as input arguments and returns a generator
+    that yields dictionaries with the values from the arrays in the input dictionary.
+
     Args:
         var_dict (dict): A dictionary where keys are variable names and values are numpy arrays.
         fun_control (dict): A dictionary which (at least) has an entry with the following key:
-            - "var_type": A list of variable types. If the entry is not "num" the corresponding
+            "var_type" (list): A list of variable types. If the entry is not "num" the corresponding
             value will be converted to the type "int".
+
     Returns:
-        dict: A dictionary with the values from the arrays in the dictionary.
-    Example:
+        Generator[dict]: A generator that yields dictionaries with the values from the arrays in the input dictionary.
+
+    Examples:
         >>> import numpy as np
+        >>> from spotPython.utils.prepare import generate_one_config_from_var_dict
         >>> var_dict = {'a': np.array([1, 3, 5]), 'b': np.array([2, 4, 6])}
         >>> fun_control = {"var_type": ["int", "num"]}
-        >>> generate_one_config_from_var_dict(var_dict, fun_control)
-        {'a': 1, 'b': 2}
+        >>> list(generate_one_config_from_var_dict(var_dict, fun_control))
+        [{'a': 1, 'b': 2}, {'a': 3, 'b': 4}, {'a': 5, 'b': 6}]
     """
     for values in iterate_dict_values(var_dict):
         values = convert_keys(values, fun_control["var_type"])
@@ -33,26 +41,31 @@ def generate_one_config_from_var_dict(var_dict, fun_control) -> dict:
         yield values
 
 
-def return_conf_list_from_var_dict(var_dict: dict, fun_control: dict) -> list:
-    """This function takes a dictionary of variables and a dictionary of function control.
-    It performs similar steps as generate_one_config_from_var_dict()
-    but returns a list of dictionaries of hyper parameter values.
+def return_conf_list_from_var_dict(
+    var_dict: Dict[str, np.ndarray], fun_control: Dict[str, Union[List[str], str]]
+) -> List[Dict[str, Union[int, float]]]:
+    """Return a list of configurations from a dictionary of variables.
+
+    This function takes a dictionary of variables and a dictionary of function control as input arguments.
+    It performs similar steps as generate_one_config_from_var_dict() but returns a list of dictionaries
+    of hyper parameter values.
+
     Args:
-        var_dict (dict): A dictionary of variables.
-        fun_control (dict): A dictionary of function control.
+        var_dict (dict): A dictionary where keys are variable names and values are numpy arrays.
+        fun_control (dict): A dictionary which (at least) has an entry with the following key:
+            "var_type" (list): A list of variable types. If the entry is not "num" the corresponding
+            value will be converted to the type "int".
+
     Returns:
-        list A list of dictionaries of hyper parameter values. Transformations are applied to the values.
+        list: A list of dictionaries of hyper parameter values. Transformations are applied to the values.
+
     Examples:
         >>> import numpy as np
-            var_dict = {'a': np.array([1]),
-                        'b': np.array([2])}
-            fun_control = {'var_type': ['int', 'int']}
-            return_conf_list_from_var_dict(var_dict, fun_control)
-            var_dict = {'a': np.array([1, 3, 5]), 'b': np.array([2, 4, 6])}
-            fun_control = {'var_type': ['int', 'int']}
-            return_conf_list_from_var_dict(var_dict, fun_control)
-            {'a': [1, 3, 5], 'b': [2, 4, 6]}
-
+        >>> from spotPython.utils.prepare import return_conf_list_from_var_dict
+        >>> var_dict = {'a': np.array([1, 3, 5]), 'b': np.array([2, 4, 6])}
+        >>> fun_control = {'var_type': ['int', 'int']}
+        >>> return_conf_list_from_var_dict(var_dict, fun_control)
+        [{'a': 1, 'b': 2}, {'a': 3, 'b': 4}, {'a': 5, 'b': 6}]
     """
     conf_list = []
     for values in generate_one_config_from_var_dict(var_dict, fun_control):
@@ -60,34 +73,33 @@ def return_conf_list_from_var_dict(var_dict: dict, fun_control: dict) -> list:
     return conf_list
 
 
-def iterate_dict_values(var_dict: dict):
-    """
-    This function takes a dictionary of variables as input arguments and returns an iterator that
-    yields the values from the arrays in the dictionary.
+def iterate_dict_values(var_dict: Dict[str, np.ndarray]) -> Generator[Dict[str, Union[int, float]], None, None]:
+    """Iterate over the values of a dictionary of variables.
 
-    Parameters:
+    This function takes a dictionary of variables as input arguments and returns a generator that
+    yields dictionaries with the values from the arrays in the input dictionary.
+
+    Args:
         var_dict (dict): A dictionary where keys are variable names and values are numpy arrays.
 
     Returns:
-        iterator: An iterator that yields the values from the arrays in the dictionary.
+        Generator[dict]:
+            A generator that yields dictionaries with the values from the arrays in the input dictionary.
 
-    Example:
+    Examples:
         >>> import numpy as np
+        >>> from spotPython.utils.prepare import iterate_dict_values
         >>> var_dict = {'a': np.array([1, 3, 5]), 'b': np.array([2, 4, 6])}
-        >>> for values in iterate_dict_values(var_dict):
-        ...     print(values)
-        {'a': 1, 'b': 2}
-        {'a': 3, 'b': 4}
-        {'a': 5, 'b': 6}
+        >>> list(iterate_dict_values(var_dict))
+        [{'a': 1, 'b': 2}, {'a': 3, 'b': 4}, {'a': 5, 'b': 6}]
     """
     n = len(next(iter(var_dict.values())))
     for i in range(n):
         yield {key: value[i] for key, value in var_dict.items()}
 
 
-def convert_keys(d: dict, var_type: list):
-    """
-    Convert values in a dictionary to integers based on a list of variable types.
+def convert_keys(d: Dict[str, Union[int, float, str]], var_type: List[str]) -> Dict[str, Union[int, float]]:
+    """Convert values in a dictionary to integers based on a list of variable types.
 
     This function takes a dictionary `d` and a list of variable types `var_type` as arguments.
     For each key in the dictionary,
@@ -96,13 +108,15 @@ def convert_keys(d: dict, var_type: list):
 
     Args:
         d (dict): The input dictionary.
-        var_type (list): A list of variable types. If the entry is not `"num"` the corresponding
-        value will be converted to the type `"int"`.
+        var_type (list):
+            A list of variable types. If the entry is not `"num"` the corresponding
+            value will be converted to the type `"int"`.
 
     Returns:
         dict: The modified dictionary with values converted to integers based on `var_type`.
 
-    Example:
+    Examples:
+        >>> from spotPython.utils.prepare import convert_keys
         >>> d = {'a': '1.1', 'b': '2', 'c': '3.1'}
         >>> var_type = ["int", "num", "int"]
         >>> convert_keys(d, var_type)
@@ -115,8 +129,9 @@ def convert_keys(d: dict, var_type: list):
     return d
 
 
-def get_dict_with_levels_and_types(fun_control, v) -> dict:
+def get_dict_with_levels_and_types(fun_control: Dict[str, Any], v: Dict[str, Any]) -> Dict[str, Any]:
     """Get dictionary with levels and types.
+
     The function maps the numerical output of the hyperparameter optimization to the corresponding levels
     of the hyperparameter needed by the core model, i.e., the tuned algorithm.
     The function takes the dictionaries fun_control and v and returns a new dictionary with the same keys as v
@@ -126,72 +141,81 @@ def get_dict_with_levels_and_types(fun_control, v) -> dict:
     If a key is not in fun_control, it takes the key from v.
     If the core_model_parameter_type value is instance, it returns the class of the value from the module
     via getattr("class", value).
-    For example,
-    if fun_control = {"HoeffdingTreeRegressor":{
-        "leaf_prediction": {
-            "levels": ["mean", "model", "adaptive"],
-            "type": "factor",
-            "default": "mean",
-            "core_model_parameter_type": "str"},
-        "leaf_model": {
-            "levels": ["linear_model.LinearRegression", "linear_model.PARegressor", "linear_model.Perceptron"],
-            "type": "factor",
-            "default": "LinearRegression",
-            "core_model_parameter_type": "instance"},
-            "splitter": {"levels": ["EBSTSplitter", "TEBSTSplitter", "QOSplitter"],
-            "type": "factor",
-            "default": "EBSTSplitter", "core_model_parameter_type": "instance()"},
-        "binary_split": {
-            "levels": [0, 1],
-            "type": "factor",
-            "default": 0,
-            "core_model_parameter_type": "bool"},
-        "stop_mem_management": {
-            "levels": [0, 1],
-            "type": "factor",
-            "default": 0,
-            "core_model_parameter_type": "bool"}}}
-        and
-        v = {'grace_period': 200,
+
+    Args:
+        fun_control (Dict[str, Any]): A dictionary containing information about the core model hyperparameters.
+        v (Dict[str, Any]): A dictionary containing the numerical output of the hyperparameter optimization.
+
+    Returns:
+        Dict[str, Any]:
+            A new dictionary with the same keys as v but with the values of the levels of the keys from fun_control.
+
+    Examples:
+        >>> fun_control = {
+        ...     "core_model_hyper_dict": {
+        ...         "leaf_prediction": {
+        ...             "levels": ["mean", "model", "adaptive"],
+        ...             "type": "factor",
+        ...             "default": "mean",
+        ...             "core_model_parameter_type": "str"
+        ...         },
+        ...         "leaf_model": {
+        ...             "levels": [
+        ...                 "linear_model.LinearRegression",
+        ...                 "linear_model.PARegressor",
+        ...                 "linear_model.Perceptron"
+        ...             ],
+        ...             "type": "factor",
+        ...             "default": "LinearRegression",
+        ...             "core_model_parameter_type": "instance"
+        ...         },
+        ...         "splitter": {
+        ...             "levels": ["EBSTSplitter", "TEBSTSplitter", "QOSplitter"],
+        ...             "type": "factor",
+        ...             "default": "EBSTSplitter",
+        ...             "core_model_parameter_type": "instance()"
+        ...         },
+        ...         "binary_split": {
+        ...             "levels": [0, 1],
+        ...             "type": "factor",
+        ...             "default": 0,
+        ...             "core_model_parameter_type": "bool"
+        ...         },
+        ...         "stop_mem_management": {
+        ...             "levels": [0, 1],
+        ...             "type": "factor",
+        ...             "default": 0,
+        ...             "core_model_parameter_type": "bool"
+        ...         }
+        ...     }
+        ... }
+        >>> v = {
+        ...     'grace_period': 200,
+        ...     'max_depth': 10,
+        ...     'delta': 1e-07,
+        ...     'tau': 0.05,
+        ...     'leaf_prediction': 0,
+        ...     'leaf_model': 0,
+        ...     'model_selector_decay': 0.95,
+        ...     'splitter': 1,
+        ...     'min_samples_split': 9,
+        ...     'binary_split': 0,
+        ...     'max_size': 500.0
+        ... }
+        >>> get_dict_with_levels_and_types(fun_control, v)
+        {
+            'grace_period': 200,
             'max_depth': 10,
             'delta': 1e-07,
             'tau': 0.05,
-            'leaf_prediction': 0,
-            'leaf_model': 0,
+            'leaf_prediction': 'mean',
+            'leaf_model': linear_model.LinearRegression,
             'model_selector_decay': 0.95,
-            'splitter': 1,
+            'splitter': TEBSTSplitter,
             'min_samples_split': 9,
-            'binary_split': 0,
-            'max_size': 500.0}
-        then the function returns
-        {'grace_period': 200,
-        'max_depth': 10,
-        'delta': 1e-07,
-        'tau': 0.05,
-        'leaf_prediction': 'mean',
-        'leaf_model': linear_model.LinearRegression,
-        'model_selector_decay': 0.95,
-        'splitter': 'TEBSTSplitter',
-        'min_samples_split': 9,
-        'binary_split': 0,
-        'max_size': 500.0}.
-
-    Args:
-        fun_control (dict): dictionary with levels and types
-        v (dict): dictionary with values
-
-    Returns:
-        new_dict (dict): dictionary with levels and types
-
-    Example:
-        >>> fun_control = {"HoeffdingTreeRegressor":{
-                "leaf_prediction": {"levels": ["mean", "model", "adaptive"],
-                                    "type": "factor",
-                                    "default": "mean",
-                                    "core_model_parameter_type": "str"}}}
-            v = {"leaf_prediction": 0}
-            get_dict_with_levels_and_types(fun_control, v)
-            {"leaf_prediction": "mean"}
+            'binary_split': False,
+            'max_size': 500.0
+        }
     """
     d = fun_control["core_model_hyper_dict"]
     new_dict = {}
@@ -219,15 +243,19 @@ def assign_values(X: np.array, var_list: list) -> dict:
     This function takes an np.array X and a list of variable names as input arguments
     and returns a dictionary where the keys are the variable names and the values are assigned from X.
 
-    Parameters:
-        X (np.array): A 2D numpy array where each column represents a variable.
-        var_list (list): A list of strings representing variable names.
+    Args:
+        X (np.array):
+            A 2D numpy array where each column represents a variable.
+        var_list (list):
+            A list of strings representing variable names.
 
     Returns:
-        dict: A dictionary where keys are variable names and values are assigned from X.
+        dict:
+            A dictionary where keys are variable names and values are assigned from X.
 
-    Example:
+    Examples:
         >>> import numpy as np
+        >>> from spotPython.utils.prepare import assign_values
         >>> X = np.array([[1, 2], [3, 4], [5, 6]])
         >>> var_list = ['a', 'b']
         >>> result = assign_values(X, var_list)
@@ -242,16 +270,22 @@ def assign_values(X: np.array, var_list: list) -> dict:
 
 def modify_hyper_parameter_levels(fun_control, hyperparameter, levels) -> dict:
     """
+    This function modifies the levels of a hyperparameter in the fun_control dictionary.
 
     Args:
-        fun_control (dict): fun_control dictionary
-        hyperparameter (str): hyperparameter name
-        levels (list): list of levels
+        fun_control (dict):
+            fun_control dictionary
+        hyperparameter (str):
+            hyperparameter name
+        levels (list):
+            list of levels
 
     Returns:
-        fun_control (dict): updated fun_control
-    Example:
+        fun_control (dict):
+            updated fun_control
+    Examples:
         >>> fun_control = {}
+            from spotPython.utils.prepare import modify_hyper_parameter_levels
             core_model  = HoeffdingTreeRegressor
             fun_control.update({"core_model": core_model})
             fun_control.update({"core_model_hyper_dict": river_hyper_dict[core_model.__name__]})
@@ -267,15 +301,20 @@ def modify_hyper_parameter_bounds(fun_control, hyperparameter, bounds) -> dict:
     """
 
     Args:
-        fun_control (dict): fun_control dictionary
-        hyperparameter (str): hyperparameter name
-        bounds (list): list of two bound values. The first value represents the lower bound
+        fun_control (dict):
+            fun_control dictionary
+        hyperparameter (str):
+            hyperparameter name
+        bounds (list):
+            list of two bound values. The first value represents the lower bound
             and the second value represents the upper bound.
 
     Returns:
-        fun_control (dict): updated fun_control
-    Example:
-        >>> fun_control = {}
+        fun_control (dict):
+            updated fun_control
+    Examples:
+        >>> from spotPython.utils.prepare import modify_hyper_parameter_levels
+            fun_control = {}
             core_model  = HoeffdingTreeRegressor
             fun_control.update({"core_model": core_model})
             fun_control.update({"core_model_hyper_dict": river_hyper_dict[core_model.__name__]})
@@ -290,11 +329,14 @@ def get_default_values(fun_control) -> dict:
     """Get the values from the "default" keys from the dictionary fun_control as a dict.
     If the key of the value has as "type" the value "int" or "float", convert the value to the corresponding type.
     Args:
-        fun_control (dict): dictionary with levels and types
+        fun_control (dict):
+            dictionary with levels and types
     Returns:
-        new_dict (dict): dictionary with default values
-    Example:
-        >>> d = {"core_model_hyper_dict":{
+        new_dict (dict):
+            dictionary with default values
+    Examples:
+        >>> from spotPython.utils.prepare import get_default_values
+            d = {"core_model_hyper_dict":{
                 "leaf_prediction": {
                     "levels": ["mean", "model", "adaptive"],
                     "type": "factor",
@@ -320,12 +362,12 @@ def get_default_values(fun_control) -> dict:
                     "type": "factor",
                     "default": 0,
                     "core_model_parameter_type": "bool"}}}
-        get_default_values_from_dict(d)
-        {'leaf_prediction': 'mean',
-        'leaf_model': 'linear_model.LinearRegression',
-        'splitter': 'EBSTSplitter',
-        'binary_split': 0,
-        'stop_mem_management': 0}
+            get_default_values_from_dict(d)
+            {'leaf_prediction': 'mean',
+            'leaf_model': 'linear_model.LinearRegression',
+            'splitter': 'EBSTSplitter',
+            'binary_split': 0,
+            'stop_mem_management': 0}
     """
     d = fun_control["core_model_hyper_dict"]
     new_dict = {}
@@ -342,11 +384,14 @@ def get_default_values(fun_control) -> dict:
 def get_var_type(fun_control) -> list:
     """Get the types of the values from the dictionary fun_control as a list.
     Args:
-        fun_control (dict): dictionary with levels and types
+        fun_control (dict):
+            dictionary with levels and types
     Returns:
-        (list): list with types
-    Example:
-        >>> d = {"core_model_hyper_dict":{
+        (list):
+            list with types
+    Examples:
+        >>> from spotPython.utils.prepare import get_var_type
+            d = {"core_model_hyper_dict":{
             "leaf_prediction": {
                 "levels": ["mean", "model", "adaptive"],
                 "type": "factor",
@@ -372,8 +417,8 @@ def get_var_type(fun_control) -> list:
                 "default": 0,
                 "core_model_parameter_type": "bool"}}}
 
-        get_var_type(d)
-        ['factor', 'factor', 'factor', 'factor', 'factor']
+            get_var_type(d)
+            ['factor', 'factor', 'factor', 'factor', 'factor']
     """
     return list(
         fun_control["core_model_hyper_dict"][key]["type"] for key in fun_control["core_model_hyper_dict"].keys()
@@ -383,11 +428,14 @@ def get_var_type(fun_control) -> list:
 def get_transform(fun_control) -> list:
     """Get the transformations of the values from the dictionary fun_control as a list.
     Args:
-        fun_control (dict): dictionary with levels and types
+        fun_control (dict):
+            dictionary with levels and types
     Returns:
-        (list): list with transformations
-    Example:
-        >>> d = {"core_model_hyper_dict":{
+        (list):
+            list with transformations
+    Examples:
+        >>> from spotPython.utils.prepare import get_transform
+            d = {"core_model_hyper_dict":{
             "leaf_prediction": {
                 "levels": ["mean", "model", "adaptive"],
                 "type": "factor",
@@ -418,8 +466,8 @@ def get_transform(fun_control) -> list:
                 "transform": "None",
                 "core_model_parameter_type": "bool"}}}
 
-        get_transform(d)
-        ['None', 'None', 'None', 'None', 'None']
+            get_transform(d)
+            ['None', 'None', 'None', 'None', 'None']
     """
     return list(
         fun_control["core_model_hyper_dict"][key]["transform"] for key in fun_control["core_model_hyper_dict"].keys()
@@ -429,10 +477,12 @@ def get_transform(fun_control) -> list:
 def get_var_name(fun_control) -> list:
     """Get the names of the values from the dictionary fun_control as a list.
     Args:
-        fun_control (dict): dictionary with names
+        fun_control (dict):
+            dictionary with names
     Returns:
-        (list): list with names
-    Example:
+        (list):
+            ist with names
+    Examples:
         >>> d = {"core_model_hyper_dict":{
             "leaf_prediction": {
                 "levels": ["mean", "model", "adaptive"],
@@ -465,20 +515,37 @@ def get_var_name(fun_control) -> list:
     return list(fun_control["core_model_hyper_dict"].keys())
 
 
-def get_bound_values(fun_control: dict, bound: str, as_list=False) -> list or np.array:
-    """Generate a list from a dictionary.
-    It takes the values from the keys "bound" in the
-    fun_control[]"core_model_hyper_dict"] dictionary and
-    returns a list of the values in the same order as the keys in the
-    dictionary.
-    For example if the dictionary is
-    {"a": {"upper": 1}, "b": {"upper": 2}}
-    the list is [1, 2] if bound="upper".
+def get_bound_values(fun_control: dict, bound: str, as_list: bool = False) -> Union[List, np.ndarray]:
+    """Generate a list or array from a dictionary.
+
+    This function takes the values from the keys "bound" in the
+    fun_control["core_model_hyper_dict"] dictionary and returns a list or array of the values
+    in the same order as the keys in the dictionary.
+
     Args:
-        fun_control (dict): dictionary with upper values
-        bound (str): either "upper" or "lower"
+        fun_control (dict):
+            A dictionary containing a key "core_model_hyper_dict"
+            which is a dictionary with keys that have either an "upper" or "lower" value.
+        bound (str):
+            Either "upper" or "lower",
+            indicating which value to extract from the inner dictionary.
+        as_list (bool):
+            If True, return a list.
+            If False, return a numpy array. Default is False.
+
     Returns:
-        (list): list with lower or upper values
+        list or np.ndarray:
+            A list or array of the extracted values.
+
+    Raises:
+        ValueError:
+            If bound is not "upper" or "lower".
+
+    Examples:
+        >>> from spotPython.utils.prepare import get_bound_values
+        >>> fun_control = {"core_model_hyper_dict": {"a": {"upper": 1}, "b": {"upper": 2}}}
+        >>> get_bound_values(fun_control, "upper", as_list=True)
+        [1, 2]
     """
     # Throw value error if bound is not upper or lower:
     if bound not in ["upper", "lower"]:
@@ -507,48 +574,53 @@ def replace_levels_with_positions(hyper_dict, hyper_dict_values) -> dict:
         "b": {"levels": ["model4", "model5", "model1"]},
         "d": {"type": "float"}},
     then the function should return {"a": 1, "b": 2, "c": 3}.
+
     Args:
-        hyper_dict (dict): dictionary with levels
-        hyper_dict_values (dict): dictionary with values
+        hyper_dict (dict):
+            dictionary with levels
+        hyper_dict_values (dict):
+            dictionary with values
     Returns:
-        (dict): dictionary with values
-    Example:
-        >>> hyper_dict = {"leaf_prediction": {
-        "levels": ["mean", "model", "adaptive"],
-        "type": "factor",
-        "default": "mean",
-        "core_model_parameter_type": "str"},
-        "leaf_model": {
-            "levels": ["linear_model.LinearRegression", "linear_model.PARegressor", "linear_model.Perceptron"],
-            "type": "factor",
-            "default": "LinearRegression",
-            "core_model_parameter_type": "instance"},
-        "splitter": {
-            "levels": ["EBSTSplitter", "TEBSTSplitter", "QOSplitter"],
-            "type": "factor",
-            "default": "EBSTSplitter",
-            "core_model_parameter_type": "instance()"},
-        "binary_split": {
-            "levels": [0, 1],
-            "type": "factor",
-            "default": 0,
-            "core_model_parameter_type": "bool"},
-        "stop_mem_management": {
-            "levels": [0, 1],
-            "type": "factor",
-            "default": 0,
-            "core_model_parameter_type": "bool"}}
-        >>> hyper_dict_values = {"leaf_prediction": "mean",
-        "leaf_model": "linear_model.LinearRegression",
-        "splitter": "EBSTSplitter",
-        "binary_split": 0,
-        "stop_mem_management": 0}
-        >>> replace_levels_with_position(hyper_dict, hyper_dict_values)
-        {'leaf_prediction': 0,
-        'leaf_model': 0,
-        'splitter': 0,
-        'binary_split': 0,
-        'stop_mem_management': 0}
+        (dict):
+            dictionary with values
+    Examples:
+        >>> from spotPython.utils.prepare import replace_levels_with_positions
+            hyper_dict = {"leaf_prediction": {
+                "levels": ["mean", "model", "adaptive"],
+                "type": "factor",
+                "default": "mean",
+                "core_model_parameter_type": "str"},
+                "leaf_model": {
+                    "levels": ["linear_model.LinearRegression", "linear_model.PARegressor", "linear_model.Perceptron"],
+                    "type": "factor",
+                    "default": "LinearRegression",
+                    "core_model_parameter_type": "instance"},
+                "splitter": {
+                    "levels": ["EBSTSplitter", "TEBSTSplitter", "QOSplitter"],
+                    "type": "factor",
+                    "default": "EBSTSplitter",
+                    "core_model_parameter_type": "instance()"},
+                "binary_split": {
+                    "levels": [0, 1],
+                    "type": "factor",
+                    "default": 0,
+                    "core_model_parameter_type": "bool"},
+                "stop_mem_management": {
+                    "levels": [0, 1],
+                    "type": "factor",
+                    "default": 0,
+                    "core_model_parameter_type": "bool"}}
+            hyper_dict_values = {"leaf_prediction": "mean",
+                "leaf_model": "linear_model.LinearRegression",
+                "splitter": "EBSTSplitter",
+                "binary_split": 0,
+                "stop_mem_management": 0}
+            replace_levels_with_position(hyper_dict, hyper_dict_values)
+                {'leaf_prediction': 0,
+                'leaf_model': 0,
+                'splitter': 0,
+                'binary_split': 0,
+                'stop_mem_management': 0}
     """
     hyper_dict_values_new = copy.deepcopy(hyper_dict_values)
     for key, value in hyper_dict_values.items():
@@ -563,10 +635,13 @@ def get_values_from_dict(dictionary) -> np.array:
     Generate an np.array that contains the values of the keys of a dictionary
     in the same order as the keys of the dictionary.
     Args:
-        dictionary (dict): dictionary with values
+        dictionary (dict):
+            dictionary with values
     Returns:
-        (np.array): array with values
-    Example:
+        (np.array):
+            array with values
+    Examples:
+        >>> from spotPython.utils.prepare import get_values_from_dict
         >>> d = {"a": 1, "b": 2, "c": 3}
         >>> get_values_from_dict(d)
         array([1, 2, 3])
@@ -577,14 +652,19 @@ def get_values_from_dict(dictionary) -> np.array:
 def add_core_model_to_fun_control(core_model, fun_control, hyper_dict, filename=None) -> dict:
     """Add the core model to the function control dictionary.
     Args:
-        core_model (class): The core model.
-        fun_control (dict): The function control dictionary.
-        hyper_dict (dict): The hyper parameter dictionary.
-        filename (str): The name of the json file that contains the hyper parameter dictionary.
-        Optional. Default is None.
+        core_model (class):
+            The core model.
+        fun_control (dict):
+            The function control dictionary.
+        hyper_dict (dict):
+            The hyper parameter dictionary.
+        filename (str):
+            The name of the json file that contains the hyper parameter dictionary.
+            Optional. Default is None.
     Returns:
-        (dict): The function control dictionary.
-    Example:
+        (dict):
+            The function control dictionary.
+    Examples:
         >>> from river.tree import HoeffdingAdaptiveTreeRegressor
             from spotRiver.data.river_hyper_dict import RiverHyperDict
             fun_control = {}
@@ -607,6 +687,27 @@ def add_core_model_to_fun_control(core_model, fun_control, hyper_dict, filename=
 
 
 def get_one_core_model_from_X(X, fun_control=None):
+    """Get one core model from X.
+    Args:
+        X (np.array):
+            The array with the hyper parameter values.
+        fun_control (dict):
+            The function control dictionary.
+    Returns:
+        (class):
+            The core model.
+    Examples:
+        >>> from river.tree import HoeffdingAdaptiveTreeRegressor
+            from spotRiver.data.river_hyper_dict import RiverHyperDict
+            fun_control = {}
+            add_core_model_to_fun_control(core_model=HoeffdingAdaptiveTreeRegressor,
+                fun_control=func_control,
+                hyper_dict=RiverHyperDict,
+                filename=None)
+            X = np.array([0, 0, 0, 0, 0])
+            get_one_core_model_from_X(X, fun_control)
+            HoeffdingAdaptiveTreeRegressor()
+    """
     var_dict = assign_values(X, fun_control["var_name"])
     config = return_conf_list_from_var_dict(var_dict, fun_control)[0]
     core_model = fun_control["core_model"](**config)
@@ -614,12 +715,58 @@ def get_one_core_model_from_X(X, fun_control=None):
 
 
 def get_one_config_from_X(X, fun_control=None):
+    """Get one config from X.
+    Args:
+        X (np.array):
+            The array with the hyper parameter values.
+        fun_control (dict):
+            The function control dictionary.
+    Returns:
+        (dict):
+            The config dictionary.
+    Examples:
+        >>> from river.tree import HoeffdingAdaptiveTreeRegressor
+            from spotRiver.data.river_hyper_dict import RiverHyperDict
+            fun_control = {}
+            add_core_model_to_fun_control(core_model=HoeffdingAdaptiveTreeRegressor,
+                fun_control=func_control,
+                hyper_dict=RiverHyperDict,
+                filename=None)
+            X = np.array([0, 0, 0, 0, 0])
+            get_one_config_from_X(X, fun_control)
+            {'leaf_prediction': 'mean',
+            'leaf_model': 'NBAdaptive',
+            'splitter': 'HoeffdingAdaptiveTreeSplitter',
+            'binary_split': 'info_gain',
+            'stop_mem_management': False}
+    """
     var_dict = assign_values(X, fun_control["var_name"])
     config = return_conf_list_from_var_dict(var_dict, fun_control)[0]
     return config
 
 
 def get_one_sklearn_model_from_X(X, fun_control=None):
+    """Get one sklearn model from X.
+    Args:
+        X (np.array):
+            The array with the hyper parameter values.
+        fun_control (dict):
+            The function control dictionary.
+    Returns:
+        (class):
+            The sklearn model.
+    Examples:
+    >>> from sklearn.linear_model import LinearRegression
+        from spotRiver.data.sklearn_hyper_dict import SklearnHyperDict
+        fun_control = {}
+        add_core_model_to_fun_control(core_model=LinearRegression,
+            fun_control=func_control,
+            hyper_dict=SklearnHyperDict,
+            filename=None)
+        X = np.array([0, 0, 0, 0, 0])
+        get_one_sklearn_model_from_X(X, fun_control)
+        LinearRegression()
+    """
     core_model = get_one_core_model_from_X(X=X, fun_control=fun_control)
     if fun_control["prep_model"] is not None:
         model = make_pipeline(fun_control["prep_model"], core_model)
@@ -629,6 +776,30 @@ def get_one_sklearn_model_from_X(X, fun_control=None):
 
 
 def get_one_river_model_from_X(X, fun_control=None):
+    """Get one river model from X.
+
+    Args:
+        X (np.array):
+            The array with the hyper parameter values.
+        fun_control (dict):
+            The function control dictionary.
+
+    Returns:
+        (class):
+            The river model.
+
+    Examples:
+        >>> from river.tree import HoeffdingAdaptiveTreeRegressor
+            from spotRiver.data.river_hyper_dict import RiverHyperDict
+            fun_control = {}
+            add_core_model_to_fun_control(core_model=HoeffdingAdaptiveTreeRegressor,
+                fun_control=func_control,
+                hyper_dict=RiverHyperDict,
+                filename=None)
+            X = np.array([0, 0, 0, 0, 0])
+            get_one_river_model_from_X(X, fun_control)
+            HoeffdingAdaptiveTreeRegressor()
+    """
     core_model = get_one_core_model_from_X(X=X, fun_control=fun_control)
     if fun_control["prep_model"] is not None:
         model = compose.Pipeline(fun_control["prep_model"], core_model)
@@ -638,6 +809,26 @@ def get_one_river_model_from_X(X, fun_control=None):
 
 
 def get_default_hyperparameters_as_array(fun_control) -> np.array:
+    """Get the default hyper parameters as array.
+    Args:
+        fun_control (dict):
+            The function control dictionary.
+
+    Returns:
+        (np.array):
+            The default hyper parameters as array.
+
+    Examples:
+        >>> from river.tree import HoeffdingAdaptiveTreeRegressor
+            from spotRiver.data.river_hyper_dict import RiverHyperDict
+            fun_control = {}
+            add_core_model_to_fun_control(core_model=HoeffdingAdaptiveTreeRegressor,
+                fun_control=func_control,
+                hyper_dict=RiverHyperDict,
+                filename=None)
+            get_default_hyperparameters_as_array(fun_control)
+            array([0, 0, 0, 0, 0])
+    """
     X0 = get_default_values(fun_control)
     X0 = replace_levels_with_positions(fun_control["core_model_hyper_dict"], X0)
     X0 = get_values_from_dict(X0)
@@ -647,8 +838,92 @@ def get_default_hyperparameters_as_array(fun_control) -> np.array:
 
 
 def get_default_hyperparameters_for_core_model(fun_control) -> dict:
+    """Get the default hyper parameters for the core model.
+    Args:
+        fun_control (dict):
+            The function control dictionary.
+
+    Returns:
+        (dict):
+            The default hyper parameters for the core model.
+
+    Examples:
+        >>> from river.tree import HoeffdingAdaptiveTreeRegressor
+            from spotRiver.data.river_hyper_dict import RiverHyperDict
+            fun_control = {}
+            add_core_model_to_fun_control(core_model=HoeffdingAdaptiveTreeRegressor,
+                fun_control=func_control,
+                hyper_dict=RiverHyperDict,
+                filename=None)
+            get_default_hyperparameters_for_core_model(fun_control)
+            {'leaf_prediction': 'mean',
+            'leaf_model': 'NBAdaptive',
+            'splitter': 'HoeffdingAdaptiveTreeSplitter',
+            'binary_split': 'info_gain',
+            'stop_mem_management': False}
+    """
     values = get_default_values(fun_control)
     values = get_dict_with_levels_and_types(fun_control=fun_control, v=values)
     values = convert_keys(values, fun_control["var_type"])
     values = transform_hyper_parameter_values(fun_control=fun_control, hyper_parameter_values=values)
     return values
+
+
+def set_data_set(fun_control, data_set) -> dict:
+    """
+    This function sets the lightning dataset in the fun_control dictionary.
+
+    Args:
+        fun_control (dict):
+            fun_control dictionary
+        data_set (class): Dataset class from torch.utils.data
+
+    Returns:
+        fun_control (dict):
+            updated fun_control
+
+    Examples:
+        >>> from spotPython.utils.init import fun_control_init
+            from spotPython.utils.prepare import set_data_module
+            from spotPython.data.lightdatamodule import LightDataModule
+            from spotPython.data.csvdataset import CSVDataset
+            from spotPython.data.pkldataset import PKLDataset
+            import torch
+            fun_control = fun_control_init()
+            ds = CSVDataset(csv_file='data.csv', target_column='prognosis', feature_type=torch.long)
+            set_data_set(fun_control=fun_control,
+                         data_set=ds)
+            fun_control["data_set"]
+    """
+    fun_control.update({"data_set": data_set})
+
+
+def set_data_module(fun_control, data_module) -> dict:
+    """
+    This function sets the lightning datamodule in the fun_control dictionary.
+
+    Args:
+        fun_control (dict):
+            fun_control dictionary
+        data_module (class): DataLoader class from torch.utils.data
+
+    Returns:
+        fun_control (dict):
+            updated fun_control
+
+    Examples:
+        >>> from spotPython.utils.init import fun_control_init
+            from spotPython.utils.prepare import set_data_module
+            from spotPython.data.lightdatamodule import LightDataModule
+            from spotPython.data.csvdataset import CSVDataset
+            from spotPython.data.pkldataset import PKLDataset
+            import torch
+            fun_control = fun_control_init()
+            dataset = CSVDataset(csv_file='data.csv', target_column='prognosis', feature_type=torch.long)
+            dm = LightDataModule(dataset=dataset, batch_size=5, test_size=7)
+            dm.setup()
+            set_data_module(fun_control=fun_control,
+                            data_module=dm)
+            fun_control["data_module"]
+    """
+    fun_control.update({"data_module": data_module})
